@@ -31,8 +31,9 @@ from version import __version__
 from cmd2 import Cmd
 
 import boto
-from boto.dynamodb.exceptions import DynamoDBResponseError, BotoClientError
+from boto.dynamodb.exceptions import DynamoDBResponseError, DynamoDBConditionalCheckFailedError, BotoClientError
 from boto.dynamodb.condition import *
+from boto.dynamodb.types import Binary
 
 import ast
 import json
@@ -113,7 +114,7 @@ class DynamoDBShell(Cmd):
     prompt = "dynash> "
 
     consistent = False
-    consumed = False
+    consumed = False  # this should really be print_consumed
     pretty = True
     verbose = False
 
@@ -150,7 +151,7 @@ class DynamoDBShell(Cmd):
         self.tables = []
         self.table = None
         self.consistent = False
-        self.print_consumed = False
+        self.consumed = False
         self.verbose = False
 
     def pprint(self, object):
@@ -176,15 +177,33 @@ class DynamoDBShell(Cmd):
     def getargs(self, line):
         return shlex.split(str(line.decode('string-escape')))
 
-    def gettype(self, stype):
-        return stype.upper()[0]
+    def get_type(self, stype):
+        if stype == 'S':
+            return "S"
+
+        if stype == 'N':
+            return 1
+
+        if stype == 'B':
+            return Binary('B')
+
+        # those are not valid key types, but anyway
+
+        if stype == 'SS':
+            return {"S"}
+
+        if stype == 'NS':
+            return {1}
+        return None
+
+    def get_first_rest(self, line):
+        param, _, rest = line.partition(" ")
+        return param, rest.strip()
 
     def get_table_params(self, line):
         if line and line[0] == ':':
-            parts = line.split(" ", 1)
-            table_name = parts[0][1:]
-            line = parts[1].strip() if len(parts) > 1 else ""
-            return self.conn.get_table(table_name), line
+            table_name, line = self.get_first_rest(line)
+            return self.conn.get_table(table_name[1:]), line
         else:
             return self.table, line
 
@@ -211,6 +230,12 @@ class DynamoDBShell(Cmd):
 
         return value
 
+    def get_list(self, line):
+        try:
+            return json.loads(line)
+        except:
+            return ast.literal_eval(line)
+
     def do_login(self, line):
         "login aws-acces-key aws-secret"
         if line:
@@ -257,8 +282,8 @@ class DynamoDBShell(Cmd):
                 rkey = ''
 
             prov = info['ProvisionedThroughput']
-            prov = " -rc:%d -wc:%d" % (prov['ReadCapacityUnits'], prov['WriteCapacityUnits'])
-            print "create %s %s%s" % (name, hkey, rkey)
+            prov = "-c %d,%d" % (prov['ReadCapacityUnits'], prov['WriteCapacityUnits'])
+            print "create %s %s %s%s" % (name, prov, hkey, rkey)
         else:
             self.pprint(desc)
 
@@ -269,28 +294,32 @@ class DynamoDBShell(Cmd):
         self.prompt = "%s> " % self.table.name
 
     def do_create(self, line):
-        "create {tablename} {hkey}[:{type} {rkey}:{type}]"
+        "create {tablename} [-c rc,wc] {hkey}[:{type} {rkey}:{type}]"
         args = self.getargs(line)
-        name = args[0]
-        hkey = args[1]
-        if ':' in hkey:
-            hkey, hkey_type = hkey.split(':')
-            hkey_type = self.gettype(hkey_type)
-        else:
-            hkey_type = self.gettype('S')
-        if len(args) > 2:
-            rkey = args[2]
-            if ':' in rkey:
-                rkey, rkey_type = rkey.split(':')
-                rkey_type = self.gettype(rkey_type)
-            else:
-                rkey_type = self.gettype('S')
+        rc = wc = 5
+
+        name = args.pop(0)  #  tablename
+
+        if args[0] == "-c": # capacity
+            args.pop(0)  # skyp -c
+
+            capacity = args.pop(0).strip()
+            rc, _, wc = capacity.partition(",")
+            rc = int(rc)
+            wc = int(wc) if wc != "" else rc
+
+        hkey, _, hkey_type = args.pop(0).partition(':')
+        hkey_type = self.get_type(hkey_type or 'S')
+
+        if args:
+            rkey, _, rkey_type = args.pop(0).partition(':')
+            rkey_type = self.get_type(rkey_type or 'S')
         else:
             rkey = rkey_type = None
 
         t = self.conn.create_table(name,
                                    self.conn.create_schema(hkey, hkey_type, rkey, rkey_type),
-                                   5, 5)
+                                   rc, wc)
         self.pprint(self.conn.describe_table(t.name))
 
     def do_drop(self, line):
@@ -343,22 +372,43 @@ class DynamoDBShell(Cmd):
                 print ""
 
     def do_put(self, line):
-        "put [:tablename] {json-body}"
+        "put [:tablename] [!fieldname:expectedvalue] {json-body} [{json-body}, {json-body}...]"
         table, line = self.get_table_params(line)
-        item = json.loads(line)
-        table.new_item(None, None, item).put()
+        expected = {}
 
-        if self.print_consumed:
-            print "consumed units:", item.consumed_units
+        while line.startswith("!"):
+            # !field:expected-value
+            param, line = self.get_first_rest(line[1:])
+            name, value = param.split(":", 1)
+            expected[name] = value or False
+
+        if line.startswith('(') or line.startswith('['):
+            list = self.get_list(line)
+            wlist = self.conn.new_batch_write_list()
+            wlist.add_batch(table, [ table.new_item(None, None, item) for item in list ])
+            response = self.conn.batch_write_item(wlist)
+            consumed = response['Responses'][table.name]['ConsumedCapacityUnits']
+
+            if 'UnprocessedItems' in response and response['UnprocessedItems']:
+                print ""
+                print "unprocessed: ", response['UnprocessedItems']
+                print ""
+        else:
+            item = json.loads(line)
+            table.new_item(None, None, item).put(expected_value=expected or None)
+            consumed = None
+
+        if self.consumed and consumed:
+            print "consumed units:", consumed
 
     def do_import(self, line):
         "import [:tablename] filename|list"
         table, line = self.get_table_params(line)
         if line[0] == '[':
-            list = ast.literal_eval(line)
+            list = self.get_list(line)
         else:
             with open(line) as f:
-                list = ast.literal_eval(f.read())
+                list = self.get_list(f.read())
 
         items = 0
         consumed = 0
@@ -372,9 +422,16 @@ class DynamoDBShell(Cmd):
         print "imported %s items, consumed units:%s" % (items, consumed)
 
     def do_update(self, line):
-        "update [:tablename] {hashkey} [-add|-delete] {attributes}"  # [ALL_OLD|ALL_NEW|UPDATED_OLD|UPDATED_NEW]"
+        "update [:tablename] [!fieldname:expectedvalue] {hashkey} [-add|-delete] {attributes}"  # [ALL_OLD|ALL_NEW|UPDATED_OLD|UPDATED_NEW]"
         table, line = self.get_table_params(line)
         hkey, attr = line.split(" ", 1)
+        expected = {}
+
+        while line.startswith("!"):
+            # !field:expected-value
+            param, line = self.get_first_rest(line[1:])
+            name, value = param.split(":", 1)
+            expected[name] = value or False
 
         if attr[0] == '-':
             op, attr = attr.split(" ", 1)
@@ -404,10 +461,10 @@ class DynamoDBShell(Cmd):
                 item.put_attribute(name, value)
 
         self.pprint(item)
-        updated = item.save(return_values=ret)
+        updated = item.save(expected_value=expected or None, return_values=ret)
         self.pprint(updated)
 
-        if self.print_consumed:
+        if self.consumed:
             print "consumed units:", item.consumed_units
 
     def do_get(self, line):
@@ -422,8 +479,7 @@ class DynamoDBShell(Cmd):
 
         if line.startswith('(') or line.startswith('[') or "," in line:
             # list of IDs
-            list = ast.literal_eval(line)
-            print ">> get %s" % str(list)
+            list = self.get_list(line)
 
             from collections import OrderedDict
 
@@ -458,7 +514,7 @@ class DynamoDBShell(Cmd):
                                   consistent_read=self.consistent)
             self.pprint(item)
 
-            if self.print_consumed:
+            if self.consumed:
                 print "consumed units:", item.consumed_units
 
     def do_rm(self, line):
@@ -493,7 +549,7 @@ class DynamoDBShell(Cmd):
                 if arg == '-c':
                     count = True
 
-                elif arg[0] == '-' and arg[1].isdigit():
+                elif arg[0] == '-' and arg[1:].isdigit():
                     max = arg[1:]
 
                 elif arg == '--':
@@ -518,7 +574,7 @@ class DynamoDBShell(Cmd):
         else:
             self.print_iterator(result)
 
-        if self.print_consumed:
+        if self.consumed:
             print "consumed units:", result.consumed_units
 
     def do_query(self, line):
@@ -535,7 +591,7 @@ class DynamoDBShell(Cmd):
             asc = True
 
         arg = args[0]
-        if arg[0] == '-' and arg[1].isdigit():
+        if arg[0] == '-' and arg[1:].isdigit():
             max = int(arg[1:])
             args.pop(0)
 
@@ -546,7 +602,7 @@ class DynamoDBShell(Cmd):
 
         self.print_iterator(result)
 
-        if self.print_consumed:
+        if self.consumed:
             print "consumed units:", result.consumed_units
 
     def do_rmall(self, line):
@@ -596,6 +652,8 @@ class DynamoDBShell(Cmd):
             return False
         except NotImplementedError as e:
             print e.message
+        except DynamoDBConditionalCheckFailedError as e:
+            print e.error_message
         except (DynamoDBResponseError, BotoClientError) as dberror:
             print self.pp.pformat(dberror)
         except:
